/* meowpow: C/C++ implementation of Meowpow, the Meowcoin Proof of Work algorithm.
 * Copyright 2019 Pawel Bylica.
 * Licensed under the Apache License, Version 2.0.
 */

#pragma once

/** The meowpow library version. */
#define MEOWPOW_VERSION "0.5.2"

#ifdef __cplusplus
namespace meowpow
{
/// The meowpow library version.
constexpr auto version = MEOWPOW_VERSION;

}  // namespace meowpow
#endif
