// meowpow: C/C++ implementation of Meowpow, the Meowcoin Proof of Work algorithm.
// Copyright 2018-2019 Pawel Bylica.
// Licensed under the Apache License, Version 2.0.

#pragma once

#include <meowpow/hash_types.h>

namespace meowpow
{
using hash256 = meowpow_hash256;
using hash512 = meowpow_hash512;
using hash1024 = meowpow_hash1024;
using hash2048 = meowpow_hash2048;
}  // namespace meowpow
